@@ -3,11 +3,13 @@ from __future__ import (absolute_import, division, print_function,
 
 import six
 import uuid
+import itertools
 import numpy as np
 import tempfile
 from io import BytesIO
 from base64 import b64encode
 from contextlib import contextmanager
+import warnings
 
 try:
     from matplotlib.colors import ColorConverter
@@ -18,11 +20,24 @@ except ImportError:
     mpl = None
     plt = None
 
+try:
+    import av
+except ImportError:
+    av = None
 
-def export(sequence, filename, rate=30, bitrate=None,
-           width=None, height=None, codec='mpeg4', format='yuv420p',
-           autoscale=True):
-    """Export a sequence of images as a standard video file.
+try:
+    with warnings.catch_warnings():
+        warnings.simplefilter("ignore")
+        from moviepy.editor import VideoClip
+except ImportError:
+    VideoClip = None
+
+
+def export_pyav(sequence, filename, rate=30, bitrate=None,
+                width=None, height=None, format=None, codec='mpeg4',
+                pixel_format='yuv420p', autoscale=None, quality=None,
+                options=None, rate_range=(16, 32)):
+    """Export a sequence of images as a standard video file using PyAv.
 
     N.B. If the quality and detail are insufficient, increase the
     bitrate.
@@ -45,30 +60,75 @@ def export(sequence, filename, rate=30, bitrate=None,
         and height is not, the height is autoscaled to maintain the aspect
         ratio.
     codec : string
-        a valid video encoding, 'mpeg4' by default
-    format: string
+        a valid video encoding, 'mpeg4' by default. Must be supported by the
+        container format. Examples are {'mpeg4', 'wmv2', 'libx264', 'rawvideo'}
+        Check https://www.ffmpeg.org/ffmpeg-codecs.html#Video-Encoders.
+    format : string
+        The container format. Guesses from the filename by default.
+    pixel_format: string
         Video stream format, 'yuv420p' by default.
+        Another possibility is 'bgr24' in combination with the 'rawvideo' codec.
     autoscale : boolean
         Linearly rescale the brightness to use the full gamut of black to
-        white values. If the datatype of the images is not 'uint8', this must
-        be set to True, as it is by default.
+        white values. False by default for uint8 readers, True otherwise.
+    quality: number or string, optional
+        For 'mpeg4' codec: sets qmin and qmax
+        For 'libx264' codec: sets crf. 0 = lossless, 23 = default.
+        For 'wmv2' codec: sets fraction of lossless bitrate, 0.01 = default
+    options : dictionary, optional
+        Dictionary that will be passed to ffmpeg. Avoid using
+        {'qscale:v', 'crf', 'pixel_format'}.
+    rate_range : tuple of two numbers
+        As extreme frame rates have playback issues on many players, by default
+        the frame rate is limited between 16 and 32. When the desired frame rate
+        is too low, frames will be multiplied an integer number of times. When
+        the desired frame rate is too high, frames will be skipped at constant
+        intervals.
 
     """
-    try:
-        import av
-    except ImportError:
+    if av is None:
         raise("This feature requires PyAV with FFmpeg or libav installed.")
-    output = av.open(filename, 'w')
-    stream = output.add_stream(bytes(codec), rate)
-    stream.pix_fmt = bytes(format)
 
-    ndim = None
-    for frame_no, img in enumerate(sequence):
-        if not frame_no:
+    export_rate = _normalize_framerate(rate, *rate_range)
+    sequence = CachedFrameGenerator(sequence, rate, autoscale)
+
+
+    # pyav is picky with unicode strings
+    codec = str(codec)
+    if format is not None:
+        format = str(format)
+    if options is not None:
+        for key in options:
+            options[str(key)] = str(options[key])
+    else:
+        options = dict()
+
+    if codec == str('wmv2') and bitrate is None and quality is None:
+        quality = 0.01
+
+    if quality is not None:
+        if codec == str('libx264'):
+            options[str('crf')] = str(quality)
+        elif codec == str('wmv2'):
+            if bitrate is not None:
+                warnings.warn("(wmv) quality is ignored when bitrate is set.")
+        elif codec == str('mpeg4'):
+            options[str('qmin')] = str(quality)
+            options[str('qmax')] = str(quality)
+        else:
+            raise NotImplemented
+
+    output = av.open(str(filename), str('w'), format=format, options=options)
+    stream = output.add_stream(codec, rate=export_rate)
+    stream.pix_fmt = str(pixel_format)
+
+    for frame_no in itertools.count():
+        try:
+            img = sequence(frame_no / export_rate)
+        except IndexError:
+            break
+        if frame_no == 0:
             # Inspect first frame to set up stream.
-            if bitrate is None:
-                bitrate = _estimate_bitrate(img.shape, rate)
-                stream.bit_rate = int(bitrate)
             if width is None:
                 stream.height = img.shape[0]
                 stream.width = img.shape[1]
@@ -76,32 +136,25 @@ def export(sequence, filename, rate=30, bitrate=None,
                 stream.width = width
                 stream.height = (height or
                                  width * img.shape[0] // img.shape[1])
-            ndim = img.ndim
 
-        if ndim == 3:
-            if img.shape.count(3) != 1:
-                raise ValueError("Images have the wrong shape.")
-            # This is a color image. Ensure that the color axis is axis 2.
-            color_axis = img.shape.index(3)
-            img = np.rollaxis(img, color_axis, 3)
-        elif ndim == 2:
-            # Expand into color to satisfy PyAV's expectation that images
-            # be in color. (Without this, an assert is tripped.)
-            img = np.repeat(np.expand_dims(img, 2), 3, axis=2)
-        else:
-            raise ValueError("Images have the wrong shape.")
+            if bitrate is not None:
+                stream.bit_rate = int(bitrate)
+            elif quality is not None and codec == str('wmv2'):
+                bitrate = quality * _estimate_bitrate([stream.height,
+                                                       stream.width],
+                                                      export_rate)
+                stream.bit_rate = int(bitrate)
 
-        # PyAV requires uint8.
-        if img.dtype is not np.uint8 and (not autoscale):
-            raise ValueError("Autoscaling must be turned on if the image "
-                             "data type is not uint8. Convert the datatype "
-                             "manually if you want to turn off autoscale.")
-        if autoscale:
-            normed = (img - img.min()) / (img.max() - img.min())
-            img = (255 * normed).astype('uint8')
-
-        frame = av.VideoFrame.from_ndarray(np.asarray(img), format=b'bgr24')
+        frame = av.VideoFrame.from_ndarray(img, format=str('rgb24'))
         packet = stream.encode(frame)
+        if packet is not None:
+            output.mux(packet)
+
+    # Finish encoding the stream
+    while True:
+        packet = stream.encode()
+        if packet is None:
+            break
         output.mux(packet)
 
     output.close()
@@ -141,13 +194,137 @@ def play(sequence, rate=30, bitrate=None,
     except ImportError:
         raise ImportError("This feature requires IPython.")
     with tempfile.NamedTemporaryFile(suffix='.webm') as temp:
-        export(sequence, bytes(temp.name), codec='libvpx', rate=rate,
-               width=width, height=height, bitrate=bitrate, format='yuv420p',
-               autoscale=True)
+        export_pyav(sequence, bytes(temp.name), codec='libvpx', rate=rate,
+                    width=width, height=height, bitrate=bitrate,
+                    format='yuv420p', autoscale=True)
         temp.flush()
         display(repr_video(temp.name, 'x-webm'))
 
 
+class CachedFrameGenerator(object):
+    def __init__(self, sequence, rate, autoscale=None, to_bgr=False):
+        self.sequence = sequence
+        self._cached_frame_no = None
+        self._cache = None
+        self.autoscale = autoscale
+        self.rate = rate
+        self.to_bgr = to_bgr
+
+    def __call__(self, t):
+        frame_no = int(t * self.rate)
+        if self._cached_frame_no != frame_no:
+            self._cached_frame_no = frame_no
+            self._cache = _to_rgb_uint8(self.sequence[frame_no], self.autoscale)
+        if self.to_bgr:
+            return self._cache[:, :, ::-1]
+        else:
+            return self._cache
+
+
+def export_moviepy(sequence, filename, rate=30, bitrate=None, width=None,
+                   height=None, codec='mpeg4', pixel_format='yuv420p',
+                   autoscale=None, quality=None, verbose=True,
+                   options=None, rate_range=(16, 32)):
+    """Export a sequence of images as a standard video file using MoviePy.
+
+    Parameters
+    ----------
+    sequence : any iterator or array of array-like images
+        The images should have two dimensions plus an
+        optional third dimensions representing color.
+    filename : string
+        name of output file
+    rate : integer, optional
+        frame rate of output file, 30 by default
+        NB: The output frame rate will be limited between `rate_range`
+    bitrate : integer or string, optional
+        Preferably use the parameter `quality` for controlling the bitrate.
+    width : integer, optional
+        By default, set the width of the images.
+    height : integer, optional
+        By default, set the  height of the images. If width is specified
+        and height is not, the height is autoscaled to maintain the aspect
+        ratio.
+    codec : string
+        a valid video encoding, 'mpeg4' by default. Must be supported by the
+        container format. Examples are {'mpeg4', 'wmv2', 'libx264', 'rawvideo'}
+        Check https://www.ffmpeg.org/ffmpeg-codecs.html#Video-Encoders.
+    pixel_format: string, optional
+        Pixel format, 'yuv420p' by default.
+        Another possibility is 'bgr24' in combination with the 'rawvideo' codec.
+    quality: number or string, optional
+        For 'mpeg4' codec: sets qscale:v. 1 = high quality, 5 = default.
+        For 'libx264' codec: sets crf. 0 = lossless, 23 = default.
+        For 'wmv2' codec: sets fraction of lossless bitrate, 0.01 = default
+    autoscale : boolean, optional
+        Linearly rescale the brightness to use the full gamut of black to
+        white values. False by default for uint8 readers, True otherwise.
+    verbose : boolean, optional
+        Determines whether MoviePy will print progress. True by default.
+    options : dictionary, optional
+        Dictionary of parameters that will be passed to ffmpeg. Avoid using
+        {'qscale:v', 'crf', 'pixel_format'}.
+    rate_range : tuple of two numbers
+        As extreme frame rates have playback issues on many players, by default
+        the frame rate is limited between 16 and 32. When the desired frame rate
+        is too low, frames will be multiplied an integer number of times. When
+        the desired frame rate is too high, frames will be skipped at constant
+        intervals.
+
+    See Also
+    --------
+    http://zulko.github.io/moviepy/ref/VideoClip/VideoClip.html#moviepy.video.VideoClip.VideoClip.write_videofile
+    """
+    if VideoClip is None:
+        raise ImportError('The MoviePy exporter requires moviepy to work.')
+
+    if options is None:
+        options = dict()
+    ffmpeg_params = []
+    for key in options:
+        ffmpeg_params.extend(['-{}'.format(key), str(options[key])])
+
+    if rate <= 0:
+        raise ValueError
+    export_rate = _normalize_framerate(rate, *rate_range)
+
+    clip = VideoClip(CachedFrameGenerator(sequence, rate, autoscale,
+                                          to_bgr=(pixel_format == 'bgr24')))
+    clip.duration = len(sequence) / rate
+    if not (height is None and width is None):
+        clip = clip.resize(height=height, width=width)
+
+    if codec == 'wmv2' and bitrate is None and quality is None:
+        quality = 0.01
+
+    if quality is not None:
+        if codec == 'libx264':
+            ffmpeg_params.extend(['-crf', str(quality)])
+        elif codec == 'mpeg4':
+            ffmpeg_params.extend(['-qscale:v', str(quality)])
+        elif codec == 'wmv2':
+            if bitrate is not None:
+                warnings.warn("(wmv) quality is ignored when bitrate is set.")
+            else:
+                bitrate = quality * _estimate_bitrate(clip.size, export_rate)
+        else:
+            raise NotImplemented
+
+    if format is not None:
+        ffmpeg_params.extend(['-pixel_format', str(pixel_format)])
+    if bitrate is not None:
+        bitrate = str(bitrate)
+
+    clip.write_videofile(filename, export_rate, codec, bitrate, audio=False,
+                         verbose=verbose, ffmpeg_params=ffmpeg_params)
+
+if av is not None:
+    export = export_pyav
+elif VideoClip is not None:
+    export = export_moviepy
+else:
+    export = None
+
 def repr_video(fname, mimetype):
     """Load the video in the file `fname`, with given mimetype,
     and display as HTML5 video.
@@ -279,12 +456,65 @@ def normalize(arr):
     return scaled_arr
 
 
+def _to_rgb_uint8(image, autoscale):
+    if autoscale is None:
+        autoscale = image.dtype != np.uint8
+
+    if autoscale:
+        image = (normalize(image) * 255).astype(np.uint8)
+    elif image.dtype != np.uint8:
+        if np.issubdtype(image.dtype, np.integer):
+            max_value = np.iinfo(image.dtype).max
+            # sometimes 12-bit images are stored as unsigned 16-bit
+            if max_value == 2**16 - 1 and image.max() < 2**12:
+                max_value = 2**12 - 1
+            image = (image / max_value * 255).astype(np.uint8)
+        else:
+            image = (image * 255).astype(np.uint8)
+
+    ndim = image.ndim
+    shape = image.shape
+    if ndim == 3 and shape.count(3) == 1:
+        # This is a color image. Ensure that the color axis is axis 2.
+        color_axis = shape.index(3)
+        image = np.rollaxis(image, color_axis, 3)
+    elif image.ndim == 3 and shape.count(4) == 1:
+        # This is an RGBA image. Drop the A values.
+        color_axis = shape.index(4)
+        image = np.rollaxis(image, color_axis, 4)[:, :, :3]
+    elif ndim == 2:
+        # Expand into color to satisfy moviepy's expectation
+        image = np.repeat(image[:, :, np.newaxis], 3, axis=2)
+    else:
+        raise ValueError("Images have the wrong shape.")
+
+    return np.asarray(image)
+
+
 def _estimate_bitrate(shape, frame_rate):
     "Return a bitrate that will guarantee lossless video."
     # Total Pixels x 8 bits x 3 channels x FPS
     return shape[0] * shape[1] * 8 * 3 * frame_rate
 
 
+def _normalize_framerate(rate, min_rate=16, max_rate=32):
+    """Limits the frame rate between min_rate and max_rate by integer multiples.
+    """
+    if rate < min_rate:
+        factor = min_rate // rate
+        if min_rate % rate > 0:
+            factor += 1
+        return rate * factor
+
+    if rate > max_rate:
+        factor = rate // max_rate
+        if rate % max_rate > 0:
+            factor += 1
+        return rate / factor
+
+    return rate
+
+
 def _monochannel_to_rgb(image, rgb):
     """This converts a greyscale image to an RGB image, using given rgb value.
 
@@ -379,7 +609,7 @@ def to_rgb(image, colors=None, normed=True):
 
 
 @contextmanager
-def _fig_size_cntx(fig, fig_size_inches, tight_layout=None):
+def _fig_size_cntx(fig, fig_size_inches, tight_layout):
     """Resize a figure in a context
 
     Parameters
@@ -389,17 +619,21 @@ def _fig_size_cntx(fig, fig_size_inches, tight_layout=None):
     fig_size_inches : tuple
         The (height, width) to use in the context. If None, the size
         is not changed
-    tight_layout : boolean or None
+    tight_layout : boolean
         When True, tight layout is used.
     """
     orig_size = fig.get_size_inches()
     orig_layout = fig.get_tight_layout()
     if fig_size_inches is not None:
         fig.set_size_inches(*fig_size_inches)
-    if tight_layout is not None:
-        fig.set_tight_layout(tight_layout)
+    fig.set_tight_layout(tight_layout)
+    if tight_layout:
+        rc_params = {'savefig.bbox': 'tight'}
+    else:
+        rc_params = {'savefig.bbox': 'standard'}
     try:
-        yield fig
+        with plt.rc_context(rc_params):
+            yield fig
     finally:
         fig.set_size_inches(*orig_size)
         fig.set_tight_layout(orig_layout)
@@ -419,7 +653,7 @@ def plot_to_frame(fig, width=512, close_fig=False, fig_size_inches=None,
         When True, the figure will be closed after plotting
     fig_size_inches : tuple
         The figure (height, width) in inches. If None, the size is not changed.
-    bbox_inches : {'tight', None}
+    bbox_inches : {None, 'standard', 'tight'}
         When 'tight', tight layout is used.
 
     Returns
@@ -435,21 +669,31 @@ def plot_to_frame(fig, width=512, close_fig=False, fig_size_inches=None,
         if fig_size_inches[0] == 0 or fig_size_inches[1] == 0:
             raise ValueError('Figure size cannot be zero.')
     if bbox_inches is None:
-        tight_layout = None
+        tight_layout = fig.get_tight_layout()
+    elif str(bbox_inches) == 'standard':
+        tight_layout = False
     elif str(bbox_inches) == 'tight':
         tight_layout = True
     else:
-        raise ValueError('Only bbox_inches=`tight` is allowed.')
+        raise ValueError("bbox_inches must be in {None, 'standard', 'tight'}")
 
     buf = BytesIO()
     with _fig_size_cntx(fig, fig_size_inches, tight_layout) as fig:
         width_in, height_in = fig.get_size_inches()
         dpi = width / width_in
-        buf_shape = (int(height_in * dpi), int(width_in * dpi), 4)
-        fig.savefig(buf, format='rgba', dpi=dpi)
+        if tight_layout:
+            # slower, but allows tight layout
+            fig.savefig(buf, format='png', dpi=dpi)
+            buf.seek(0)
+            image = plt.imread(buf)
+        else:
+            # faster, but only possible without tight layout
+            fig.savefig(buf, format='rgba', dpi=dpi)
+            buf.seek(0)
+            buf_shape = (int(height_in * dpi), int(width_in * dpi), 4)
+            image = np.fromstring(buf.read(),
+                                  dtype='uint8').reshape(*buf_shape)
 
-    buf.seek(0)
-    image = np.fromstring(buf.read(), dtype='uint8').reshape(*buf_shape)
     if close_fig:
         plt.close(fig)
     return Frame(image)
